@@ -1,6 +1,7 @@
+{-# LANGUAGE CPP #-}
 {-|
 Module      : HsLua.Module.System
-Copyright   : © 2019-2024 Albert Krewinkel
+Copyright   : © 2019-2025 Albert Krewinkel
 License     : MIT
 Maintainer  : Albert Krewinkel <tarleb@hslua.org>
 Stability   : alpha
@@ -20,23 +21,31 @@ module HsLua.Module.System (
   , os
 
   -- ** Functions
+  , cmd
+  , cp
   , cputime
   , env
   , getenv
   , getwd
   , ls
   , mkdir
+  , read_file
+  , rename
+  , rm
   , rmdir
   , setenv
   , setwd
+  , times
   , tmpdirname
   , with_env
   , with_tmpdir
   , with_wd
+  , write_file
+  , xdg
   )
 where
 
-import Control.Monad (forM_)
+import Control.Monad ((>=>), forM_)
 import Control.Monad.Catch (bracket)
 import Data.Maybe (fromMaybe)
 import Data.Text (Text)
@@ -46,13 +55,18 @@ import HsLua.Marshalling
 import HsLua.Packaging
 import HsLua.Module.SystemUtils
 
+import qualified Data.ByteString as B
 import qualified Data.Text as T
-import qualified HsLua.Core as Lua
+import qualified Data.Time as Time
+import qualified Data.Time.Format.ISO8601 as ISO8601
+import qualified HsLua.Core.Utf8 as Utf8
 import qualified System.CPUTime as CPUTime
 import qualified System.Directory as Directory
 import qualified System.Environment as Env
+import qualified System.Exit as Exit
 import qualified System.Info as Info
 import qualified System.IO.Temp as Temp
+import qualified System.Process as Process
 
 -- | The "system" module.
 documentedModule :: LuaError e => Module e
@@ -66,19 +80,27 @@ documentedModule = Module
       , os
       ]
   , moduleFunctions =
-      [ cputime
+      [ cmd
+      , cp
+      , cputime
       , env
       , getenv
       , getwd
       , ls
       , mkdir
+      , read_file
+      , rename
+      , rm
       , rmdir
       , setenv
       , setwd
+      , times
       , tmpdirname
       , with_env
       , with_tmpdir
       , with_wd
+      , write_file
+      , xdg
       ]
   , moduleOperations = []
   , moduleTypeInitializers = []
@@ -144,7 +166,11 @@ os :: Field e
 os = Field
   { fieldName = "os"
   , fieldType = "string"
-  , fieldDescription = "The operating system on which the program is running."
+  , fieldDescription = T.unlines
+    [ "The operating system on which the program is running."
+    , "The most common values are `darwin` (macOS), `freebsd`, `linux`,"
+    , "`linux-android`, `mingw32` (Windows), `netbsd`, `openbsd`."
+    ]
   , fieldPushValue = pushString Info.os
   }
 
@@ -153,6 +179,41 @@ os = Field
 -- Functions
 --
 
+-- | Run a system command
+cmd :: LuaError e => DocumentedFunction e
+cmd = defun "cmd"
+  ### (\command args minput opts -> do
+          let input = fromMaybe "" minput
+          let cp_opts = (Process.proc command args)
+                        { Process.env = processOptsEnv =<< opts
+                        , Process.cwd = processOptsCwd =<< opts
+                        }
+          liftIO $ Process.readCreateProcessWithExitCode cp_opts input)
+  <#> filepathParam "command" "command to execute"
+  <#> parameter (peekList peekString) "{string,...}" "args" "command arguments"
+  <#> opt (parameter peekString "string" "input" "input on stdin")
+  <#> opt (parameter peekProcessOptions "table" "opts" "process options")
+  =#> (functionResult (pushExitCode . (\(a,_,_) -> a)) "integer|boolean"
+                      "exit code – `false` on success, an integer otherwise" <>
+       functionResult (pushString . \(_,b,_) -> b) "string" "stdout" <>
+       functionResult (pushString . \(_,_,c) -> c) "string" "stderr")
+  #? T.unlines
+     [ "Executes a system command with the given arguments and `input`"
+     , "on *stdin*."
+     ]
+
+-- | Copy a file
+cp :: LuaError e => DocumentedFunction e
+cp = defun "cp"
+  ### (\src tgt -> ioToLua $ Directory.copyFile src tgt)
+  <#> filepathParam "source" "source file"
+  <#> filepathParam "target" "target destination"
+  =#> []
+  #? T.unlines
+     [ "Copy a file with its permissions."
+     , "If the destination file already exists, it is overwritten."
+     ]
+
 -- | Access the CPU time, e.g. for benchmarking.
 cputime :: LuaError e => DocumentedFunction e
 cputime = defun "cputime"
@@ -230,6 +291,48 @@ mkdir = defun "mkdir"
        , "created as necessary.\n"
        ]
 
+-- | Returns the contents of a file.
+read_file :: LuaError e => DocumentedFunction e
+read_file = defun "read_file"
+  ### (ioToLua . B.readFile)
+  <#> filepathParam "filepath" "File to read"
+  =#> functionResult pushByteString "string" "file contents"
+
+-- | Rename a file path.
+rename :: LuaError e => DocumentedFunction e
+rename = defun "rename"
+  ### (\old new -> ioToLua $ do
+          isDir <- Directory.doesDirectoryExist old
+          if isDir
+            then Directory.renameDirectory old new
+            else Directory.renameFile old new)
+  <#> filepathParam "old" "original path"
+  <#> filepathParam "new" "new path"
+  =#> []
+  #? T.unlines
+     [ "Change the name of an existing path from `old` to `new`."
+     , ""
+     , "If `old` is a directory and `new` is a directory that already"
+     , "exists, then `new` is atomically replaced by the `old` directory."
+     , "On Win32 platforms, this function fails if `new` is an existing"
+     , "directory."
+     , ""
+     , "If `old` does not refer to a directory, then neither may `new`."
+     , ""
+     , "Renaming may not work across file system boundaries or due to"
+     , "other system-specific reasons.  It's generally more robust to"
+     , "copy the source path to its destination before deleting the"
+     , "source."
+     ]
+
+-- | Remove a file.
+rm :: LuaError e => DocumentedFunction e
+rm = defun "rm"
+  ### ioToLua . Directory.removeFile
+  <#> filepathParam "filename" "file to remove"
+  =#> []
+  #? "Removes the directory entry for an existing file."
+
 -- | Remove an existing directory.
 rmdir :: LuaError e => DocumentedFunction e
 rmdir = defun "rmdir"
@@ -261,6 +364,22 @@ setwd = defun "setwd"
   =#> []
   #? "Change the working directory to the given path."
 
+-- | Get the modification time and access time of a file.
+times :: LuaError e => DocumentedFunction e
+times = defun "times"
+  ### (\filepath -> ioToLua $
+       (,) <$> Directory.getModificationTime filepath
+           <*> Directory.getAccessTime filepath)
+  <#> filepathParam "filepath" "file or directory path"
+  =#> (functionResult (pushUTCTime . fst) "table"
+       "time at which the file or directory was last modified" <>
+       functionResult (pushUTCTime . snd) "table"
+       "time at which the file or directory was last accessed")
+  #? T.unlines
+     [ "Obtain the modification and access time of a file or directory."
+     , "The times are returned as strings using the ISO 8601 format."
+     ]
+
 -- | Get the current directory for temporary files.
 tmpdirname :: LuaError e => DocumentedFunction e
 tmpdirname = defun "tmpdirname"
@@ -291,10 +410,10 @@ tmpdirname = defun "tmpdirname"
 with_wd :: LuaError e => DocumentedFunction e
 with_wd = defun "with_wd"
   ### (\fp callback ->
-        bracket (Lua.liftIO Directory.getCurrentDirectory)
-                (Lua.liftIO . Directory.setCurrentDirectory)
-        (\_ -> do
-              Lua.liftIO (Directory.setCurrentDirectory fp)
+        bracket (ioToLua Directory.getCurrentDirectory)
+                (ioToLua . Directory.setCurrentDirectory)
+           (\_ -> do
+              ioToLua (Directory.setCurrentDirectory fp)
               callback `invokeWithFilePath` fp))
   <#> filepathParam "directory"
         "Directory in which the given `callback` should be executed"
@@ -312,7 +431,7 @@ with_wd = defun "with_wd"
 with_env :: LuaError e => DocumentedFunction e
 with_env = defun "with_env"
   ### (\environment callback ->
-        bracket (Lua.liftIO Env.getEnvironment)
+        bracket (ioToLua Env.getEnvironment)
                 setEnvironment
                 (\_ -> setEnvironment environment *> invoke callback))
   <#> parameter (peekKeyValuePairs peekString peekString) "table"
@@ -330,13 +449,14 @@ with_env = defun "with_env"
      , "action."
      ]
  where
-  setEnvironment newEnv = Lua.liftIO $ do
+  setEnvironment newEnv = ioToLua $ do
     -- Crude, but fast enough: delete all entries in new environment,
     -- then restore old environment one-by-one.
     curEnv <- Env.getEnvironment
     forM_ curEnv (Env.unsetEnv . fst)
     forM_ newEnv (uncurry Env.setEnv)
 
+-- | Provides a temporary directory for the given action.
 with_tmpdir :: LuaError e => DocumentedFunction e
 with_tmpdir = defun "with_tmpdir"
   ### (\mParentDir tmpl callback -> case mParentDir of
@@ -371,6 +491,54 @@ with_tmpdir = defun "with_tmpdir"
           return Nothing
         else Just <$> peekString idx
 
+-- | Write a string to a file.
+write_file :: LuaError e => DocumentedFunction e
+write_file = defun "write_file"
+  ### (\filepath contents ->
+         ioToLua $ B.writeFile filepath contents)
+  <#> filepathParam "filepath" "path to target file"
+  <#> parameter peekByteString "string" "contents" "file contents"
+  =#> []
+  #? "Writes a string to a file."
+
+-- | Obtain the paths to special directories.
+xdg :: LuaError e => DocumentedFunction e
+xdg = defun "xdg"
+  ### (\xdgDirTypeOrList mfp->
+         case xdgDirTypeOrList of
+           Left xdgDirType -> Left <$>
+             let fp = fromMaybe "" mfp
+             in ioToLua $ Directory.getXdgDirectory xdgDirType fp
+           Right xdgDirList ->
+             ioToLua $ Right <$> Directory.getXdgDirectoryList xdgDirList)
+
+  <#> parameter peekXdgDirectory "string" "xdg_directory_type"
+        (T.unlines
+         [ "The type of the XDG directory or search path."
+         , "Must be one of `config`, `data`, `cache`, `state`,"
+         , "`datadirs`, or `configdirs`."
+         , ""
+         , "Matching is case-insensitive, and underscores and `XDG`"
+         , "prefixes are ignored, so a value like"
+         , "`XDG_DATA_DIRS` is also acceptable."
+         , ""
+         , "The `state` directory might not be available, depending"
+         , "on the version of the underlying Haskell library."
+         ])
+  <#> opt (filepathParam "filepath"
+           ("relative path that is appended to the path; ignored " <>
+            "if the result is a list of search paths."))
+  =#> functionResult (either pushString (pushList pushString))
+        "string|{string,...}"
+        "Either a single file path, or a list of search paths."
+  #? T.unlines
+     [ "Access special directories and directory search paths."
+     , ""
+     , "Special directories for storing user-specific application"
+     , "data, configuration, and cache files, as specified by the"
+     , "[XDG Base Directory Specification](" <>
+       "https://specifications.freedesktop.org/basedir-spec/latest/)."
+     ]
 
 --
 -- Parameters
@@ -386,3 +554,62 @@ filepathParam = stringParam
 filepathResult :: Text -- ^ Description
                -> [FunctionResult e FilePath]
 filepathResult = functionResult pushString "string"
+
+--
+-- Process parameters
+--
+
+-- | Process options
+data ProcessOpts = ProcessOpts
+  { processOptsEnv  :: Maybe [(String, String)]
+  , processOptsCwd  :: Maybe FilePath
+  }
+
+-- | Peek process creation options
+peekProcessOptions :: LuaError e => Peeker e ProcessOpts
+peekProcessOptions = typeChecked "table" istable $ \idx -> do
+  let peekEnv = peekKeyValuePairs peekString peekString
+  env' <- peekFieldRaw (peekNilOr peekEnv)    "env" idx
+  cwd' <- peekFieldRaw (peekNilOr peekString) "cwd" idx
+  return $ ProcessOpts
+    { processOptsEnv = env'
+    , processOptsCwd = cwd'
+    }
+
+-- | Pushes an exit code; failure codes are pushed as integers, and
+-- success is pushed as `false`. This means that the value can be
+-- interpreted as a boolean `failed` value.
+pushExitCode :: Pusher e Exit.ExitCode
+pushExitCode = \case
+  Exit.ExitSuccess   -> pushBool False
+  Exit.ExitFailure n -> pushIntegral n
+
+-- | Pushes a time as ISO 8601 string.
+pushUTCTime :: Pusher e Time.UTCTime
+pushUTCTime = pushString . ISO8601.iso8601Show
+
+-- | Get an XDG directory type identifier.
+peekXdgDirectory :: Peeker e
+                    (Either Directory.XdgDirectory Directory.XdgDirectoryList)
+peekXdgDirectory =
+  (fmap cleanupXdgSpec . peekText) >=> \case
+    "cache"       -> pure (Left Directory.XdgCache)
+    "config"      -> pure (Left Directory.XdgConfig)
+    "data"        -> pure (Left Directory.XdgData)
+#if MIN_VERSION_directory(1,3,7)
+    "state"       -> pure (Left Directory.XdgState)
+#endif
+    "datadirs"    -> pure (Right Directory.XdgDataDirs)
+    "configdirs"  -> pure (Right Directory.XdgConfigDirs)
+    s             -> failPeek $
+      "Expected 'cache', 'config', 'data', or 'state', got: " <>
+      Utf8.fromText s
+  where
+    -- Cleanup the XDG directory specifier as to make matching easier
+    -- while keeping things permissive.
+    -- Remove underscores, any 'xdg' prefix, and
+    -- make sure everything is lowercase
+    cleanupXdgSpec =
+      (\s -> fromMaybe s $ T.stripPrefix "xdg" s)
+      . T.filter (/= '_')
+      . T.toLower
